# Copyright 2015-2016 Canonical Ltd.
#
# This file is part of the Telegraf Charm for Juju.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import base64
import binascii
import io
import os
import json
import time
import yaml

from charms.reactive import (
    helpers,
    hook,
    when,
    when_not,
    set_state,
    remove_state,
    toggle_state,
)
from charms.reactive.bus import get_states

from charmhelpers import context
from charmhelpers.core import hookenv, host, unitdata
from charmhelpers.core.templating import render
from charmhelpers.fetch import apt_cache
import charms.promreg

from jinja2 import Template

BASE_DIR = '/etc/telegraf'

CONFIG_FILE = 'telegraf.conf'

CONFIG_DIR = 'telegraf.d'


# Utilities #
def exec_timeout_supported():
    cache = apt_cache()
    pkg = cache['telegraf']
    timeout_support = True
    if '0.12' in pkg.current_ver.ver_str:
        timeout_support = False
    return timeout_support


def get_templates_dir():
    return os.path.join(hookenv.charm_dir(), 'templates')


def get_main_config_path():
    return os.path.join(BASE_DIR, CONFIG_FILE)


def get_configs_dir():
    return os.path.join(BASE_DIR, CONFIG_DIR)


def list_supported_plugins():
    return [k for k in hookenv.metadata()['requires'].keys()
            if k != 'juju-info'] + \
        [k for k in hookenv.metadata()['provides'].keys()
            if k != 'juju-info']


def list_config_files():
    config_files = [get_main_config_path()]
    # only include config files for configured plugins
    current_states = get_states()
    for plugin in list_supported_plugins():
        if 'plugins.{}.configured'.format(plugin) in current_states.keys():
            config_path = '{}/{}.conf'.format(get_configs_dir(), plugin)
            config_files.append(config_path)
    if 'extra_plugins.configured' in current_states.keys():
        config_files.append('{}/extra_plugins.conf'.format(get_configs_dir()))
    return config_files


def get_hostname_label():
    config = hookenv.config()
    hostname_fmt = config['hostname']
    unit = get_remote_unit_name().replace('/', '-')  # / is invalid in labels.
    if hostname_fmt == 'UNIT_NAME':  # Deprecated
        return unit
    env = os.environ
    model = env.get('JUJU_ENV_NAME') or env.get('JUJU_MODEL_NAME', '')
    uuid = env.get('JUJU_ENV_UUID') or env.get('JUJU_MODEL_UUID', '')
    return hostname_fmt.format(unit=unit, model=model, uuid=uuid)


def get_remote_unit_name():
    for rel_type in hookenv.metadata()['requires'].keys():
        rels = hookenv.relations_of_type(rel_type)
        if rels and len(rels) >= 1:
            rel = rels[0]
            if rel['private-address'] == hookenv.unit_private_ip():
                return rel['__unit__']


def render_base_inputs():
    extra_options = get_extra_options()
    # use base inputs from charm templates
    with open(os.path.join(get_templates_dir(), 'base_inputs.conf'), 'r') as fd:
        return render_template(
            fd.read(),
            {'extra_options': extra_options['inputs']})


def get_extra_options():
    extra_options = {'inputs': {}, 'outputs': {}}
    extra_options_raw = hookenv.config()['extra_options']
    extra_opts = yaml.load(extra_options_raw) or {}
    extra_options.update(extra_opts)
    # jsonify value, required as the telegraf config values format is similar
    # to raw json
    json_vals = {}
    # kind level
    for k, v in extra_options.items():
        json_vals[k] = {}
        # plugins level
        for plugin, values in v.items():
            json_vals[k][plugin] = {}
            # inner plugin (aka key:value)
            for key, val in values.items():
                if key in ('tagpass', 'tagdrop'):
                    # this is a tagpass/drop, we need to go deeper
                    json_vals[k][plugin][key] = {}
                    for tag, tagvalue in val.items():
                        json_vals[k][plugin][key][tag] = json.dumps(tagvalue)
                else:
                    json_vals[k][plugin][key] = json.dumps(val)
    return json_vals


def render_extra_options(kind, name, extra_options=None):
    template = """
  {% if extra_options %}
  {% for key, value in extra_options.items() %}
  {% if key != 'tagpass' and key != 'tagdrop' %}
  {{ key }} = {{ value }}
  {% endif %}
  {% endfor %}
  {% for key, value in extra_options.items() %}
  {% if key == 'tagpass' or key == 'tagdrop' %}
  [{{ kind }}.{{ name }}.{{ key }}]
    {% for tag, tagvalue in value.items() %}
    {{ tag }} = {{ tagvalue }}
    {% endfor %}
  {% endif %}
  {% endfor %}
  {% endif %}
  """
    if extra_options is None:
        extra_options = get_extra_options()
    context = {"extra_options": extra_options[kind].get(name, {}),
               "kind": kind,
               "name": name}
    return render_template(template, context)


def render_template(template, context):
    tmpl = Template(template, lstrip_blocks=True, trim_blocks=True)
    return tmpl.render(**context)


def check_prometheus_port(key, new_port):
    unitdata_key = '{}.port'.format(key)
    kv = unitdata.kv()
    existing_port = kv.get(unitdata_key)
    if existing_port != new_port:
        hookenv.open_port(new_port)
        if existing_port:  # Dont try to close non existing ports
            hookenv.close_port(existing_port)
            charms.promreg.deregister(None, existing_port)
        kv.set(unitdata_key, new_port)

    # Register the port, even if this is a re-registration
    charms.promreg.register(None, new_port)


def get_prometheus_port():
    config = hookenv.config()
    if not config.get('prometheus_output_port', False):
        return False
    if config.get('prometheus_output_port') == 'default':
        return 9103
    else:
        return int(config.get('prometheus_output_port'))


# States


@when_not('telegraf.configured')
def configure_telegraf():
    config_path = get_main_config_path()
    if get_remote_unit_name() is None:
        hookenv.status_set('waiting', 'Waiting for juju-info relation')
        # if UNIT_NAME in hostname config and relation not yet available,
        # make telegraf unable to start to not get weird metrics names
        if os.path.exists(config_path):
            os.unlink(config_path)
        return

    config = hookenv.config()
    context = config.copy()
    inputs = config.get('inputs_config', '')
    outputs = config.get('outputs_config', '')
    # just for the migration out of base64
    if inputs:
        try:
            inputs = base64.b64decode(inputs.encode('utf-8'), validate=True).decode('utf-8')
        except binascii.Error:
            # not bas64, probably already up to date configs
            pass
    if outputs:
        try:
            outputs = base64.b64decode(outputs.encode('utf-8'), validate=True).decode('utf-8')
        except binascii.Error:
            # not bas64, probably already up to date configs
            pass
    tags = []
    if config['tags']:
        for tag in config['tags'].split(','):
            key, value = tag.split("=")
            tags.append('{} = "{}"'.format(key, value))
    context["tags"] = tags
    if inputs:
        context["inputs"] = inputs
    else:
        # use base inputs from charm templates
        context["inputs"] = render_base_inputs()
    if outputs:
        context["outputs"] = outputs
    else:
        context["outputs"] = ""
        hookenv.log("No output plugins in main config.")

    context["hostname"] = get_hostname_label()

    if get_prometheus_port():
        context["prometheus_output_port"] = get_prometheus_port()
        check_prometheus_port("prometheus_output", get_prometheus_port())
    context['extra_options'] = get_extra_options()

    hookenv.log("Updating main config file")
    render(source='telegraf.conf.tmpl', templates_dir=get_templates_dir(),
           target=config_path, context=context)
    set_state('telegraf.configured')
    set_state('telegraf.needs_reload')


@hook('upgrade-charm')
def upgrade_charm():
    for plugin in list_supported_plugins():
        remove_state('plugins.{}.configured'.format(plugin))
    remove_state('extra_plugins.configured')
    remove_state('telegraf.configured')


@when('config.changed')
def handle_config_changes():
    config = hookenv.config()
    if config.changed('extra_options'):
        for plugin in list_supported_plugins():
            remove_state('plugins.{}.configured'.format(plugin))
    # if something else changed, let's reconfigure telegraf itself just in case
    if config.changed('extra_plugins'):
        remove_state('extra_plugins.configured')
    remove_state('telegraf.configured')


@when('telegraf.configured')
@when_not('extra_plugins.configured')
def configure_extra_plugins():
    config = hookenv.config()
    plugins = config['extra_plugins']
    if plugins:
        config_path = '{}/extra_plugins.conf'.format(get_configs_dir())
        host.write_file(config_path, plugins.encode('utf-8'))
        set_state('extra_plugins.configured')
        set_state('telegraf.needs_reload')


@when('elasticsearch.available')
def elasticsearch_input(es):
    template = """
[[inputs.elasticsearch]]
  servers = {{ servers }}
"""
    hosts = []
    rels = hookenv.relations_of_type('elasticsearch')
    for rel in rels:
        es_host = rel.get('host')
        port = rel.get('port')
        if not es_host or not port:
            hookenv.log('No host received for relation: {}.'.format(rel))
            continue
        hosts.append("http://{}:{}".format(es_host, port))
    config_path = '{}/{}.conf'.format(get_configs_dir(), 'elasticsearch')
    if hosts:
        context = {"servers": json.dumps(hosts)}
        input_config = render_template(template, context) + \
            render_extra_options("inputs", "elasticsearch")
        hookenv.log("Updating {} plugin config file".format('elasticsearch'))
        host.write_file(config_path, input_config.encode('utf-8'))
        set_state('plugins.elasticsearch.configured')
    elif os.path.exists(config_path):
        os.unlink(config_path)
        remove_state('plugins.elasticsearch.configured')
    set_state('telegraf.needs_reload')


@when('memcached.available')
def memcached_input(memcache):
    template = """
[[inputs.memcached]]
  servers = {{ servers }}
"""
    required_keys = ['host', 'port']
    rels = hookenv.relations_of_type('memcached')
    addresses = []
    for rel in rels:
        if all([rel.get(key) for key in required_keys]):
            addr = rel['host']
            port = rel['port']
            address = '{}:{}'.format(addr, port)
            addresses.append(address)
    config_path = '{}/{}.conf'.format(get_configs_dir(), 'memcached')
    if addresses:
        context = {"servers": json.dumps(addresses)}
        input_config = render_template(template, context) + \
            render_extra_options("inputs", "memcached")
        hookenv.log("Updating {} plugin config file".format('memcached'))
        host.write_file(config_path, input_config.encode('utf-8'))
        set_state('plugins.memcached.configured')
    elif os.path.exists(config_path):
        os.unlink(config_path)
    set_state('telegraf.needs_reload')


@when('mongodb.database.available')
def mongodb_input(mongodb):
    template = """
[[inputs.mongodb]]
  servers = {{ servers }}
"""
    rels = hookenv.relations_of_type('mongodb')
    mongo_addresses = []
    for rel in rels:
        addr = rel['private-address']
        port = rel.get('port', None)
        if port:
            mongo_address = '{}:{}'.format(addr, port)
        else:
            mongo_address = addr
        mongo_addresses.append(mongo_address)
    config_path = '{}/{}.conf'.format(get_configs_dir(), 'mongodb')
    if mongo_addresses:
        context = {"servers": json.dumps(mongo_addresses)}
        input_config = render_template(template, context) + \
            render_extra_options("inputs", "mongodb")
        hookenv.log("Updating {} plugin config file".format('mongodb'))
        host.write_file(config_path, input_config.encode('utf-8'))
        set_state('plugins.mongodb.configured')
    elif os.path.exists(config_path):
        os.unlink(config_path)
    set_state('telegraf.needs_reload')


@when('mysql.available')
def mysql_input(mysql):
    contexts = []
    for relid, relation in context.Relations()[mysql.relation_name].items():
        for unit, reldata in relation.items():
            hookenv.log("Available relations", level="DEBUG")
            if reldata['private-address'] == hookenv.unit_private_ip():
                if mysql.connection_string():
                    contexts.append({'host': mysql.host(),
                                     'port': mysql.port(),
                                     'user': mysql.user(),
                                     'pass': mysql.password(),
                                     'is_secure': 'false',   # TODO: provide config intf for this
                                     'slave': reldata.get('slave', None)})
                    break

    render_mysql_tmpl(contexts)
    toggle_state('plugins.mysql.configured', bool(contexts))
    set_state('telegraf.needs_reload')


@when('postgresql.database.connected')
def choose_postgresql_database(pgsql):
    # We have no need for our own database. If we need to create
    # custom views, we can remove this handler. Or maybe it should
    # be configurable, so people can use telegraf to export custom
    # metrics about their data.
    pgsql.set_database('postgres')


@when('postgresql.database.available')
def postgresql_input(pgsql):
    # The subordinate may be connected to several services.
    contexts = []
    for relid, relation in context.Relations()[pgsql.relation_name].items():
        for unit, reldata in relation.items():
            if reldata['private-address'] == hookenv.unit_private_ip():
                conn_str = pgsql.connection_string(unit)
                if conn_str:
                    contexts.append({'conn_str': conn_str,
                                     'server': conn_str.host,
                                     'replica': reldata['state']})
                    break
        else:
            hookenv.log('Local database not yet available on {}'.format(relid))
            break

    render_postgresql_tmpl(contexts)
    toggle_state('plugins.postgresql.configured', bool(contexts))
    set_state('telegraf.needs_reload')


def render_mysql_tmpl(contexts):
    config_path = '{}/{}.conf'.format(get_configs_dir(), 'mysql')

    if contexts:
        f = io.StringIO()
        template = open(os.path.join(get_templates_dir(),
                                     'mysql.tmpl'), 'r').read()
        for ctx in contexts:
            f.write(render_template(template, ctx))
            f.write(render_extra_options("inputs", "mysql"))
        host.write_file(config_path, f.getvalue().encode('UTF-8'))
    elif os.path.exists(config_path):
        os.unlink(config_path)


def render_postgresql_tmpl(contexts):
    config_path = '{}/{}.conf'.format(get_configs_dir(), 'postgresql')

    if contexts:
        f = io.StringIO()
        template = open(os.path.join(get_templates_dir(),
                                     'postgresql.tmpl'), 'r').read()
        for ctx in contexts:
            f.write(render_template(template, ctx))
            f.write(render_extra_options("inputs", "postgresql"))
        host.write_file(config_path, f.getvalue().encode('UTF-8'))
    elif os.path.exists(config_path):
        os.unlink(config_path)


@when('haproxy.available')
def haproxy_input(haproxy):
    template = """
[[inputs.haproxy]]
  servers = {{ servers }}
"""
    rels = hookenv.relations_of_type('haproxy')
    haproxy_addresses = []
    for rel in rels:
        enabled = rel.get('enabled', False)
        # Juju gives us a string instead of a boolean, fix it
        if isinstance(enabled, str):
            if enabled in ['y', 'yes', 'true', 't', 'on', 'True']:
                enabled = True
            else:
                enabled = False
        if not enabled:
            continue
        addr = rel['private-address']
        if addr == hookenv.unit_private_ip():
            addr = "localhost"
        port = rel['port']
        user = rel['user']
        password = rel.get('password', None)
        userpass = user
        if password:
            userpass += ":{}".format(password)
        haproxy_address = 'http://{}@{}:{}'.format(userpass, addr, port)
        haproxy_addresses.append(haproxy_address)
    config_path = '{}/{}.conf'.format(get_configs_dir(), 'haproxy')
    if haproxy_addresses:
        input_config = render_template(template, {"servers": json.dumps(haproxy_addresses)}) + \
            render_extra_options("inputs", "haproxy")
        hookenv.log("Updating {} plugin config file".format('haproxy'))
        host.write_file(config_path, input_config.encode('utf-8'))
        set_state('plugins.haproxy.configured')
    elif os.path.exists(config_path):
        os.unlink(config_path)
    set_state('telegraf.needs_reload')


@when('apache.available')
def apache_input(apache):
    template = """
[[inputs.apache]]
  urls = {{ urls }}
"""
    config_path = '{}/{}.conf'.format(get_configs_dir(), 'apache')
    port = '8080'
    vhost = render(source='apache-server-status.tmpl',
                   templates_dir=get_templates_dir(),
                   target=None,
                   context={'port': port})
    relation_info = {"ports": port,
                     "domain": "apache-status",
                     "enabled": True,
                     "site_config": vhost,
                     "site_modules": "status"}
    urls = []
    rels = hookenv.relations_of_type('apache')
    for rel in rels:
        hookenv.relation_set(rel['__relid__'], relation_settings=relation_info)
        addr = rel['private-address']
        url = 'http://{}:{}/server-status?auto'.format(addr, port)
        urls.append(url)
    if urls:
        context = {"urls": json.dumps(urls)}
        input_config = render_template(template, context) + \
            render_extra_options("inputs", "apache")
        hookenv.log("Updating {} plugin config file".format('apache'))
        host.write_file(config_path, input_config.encode('utf-8'))
        set_state('plugins.apache.configured')
    elif os.path.exists(config_path):
        os.unlink(config_path)
    set_state('telegraf.needs_reload')


@when('exec.available')
def exec_input(exec_rel):
    template = """
{% for cmd in commands %}
[[inputs.exec]]
  commands = {{ cmd.commands }}
  {% for key, value in cmd|dictsort %}
      {% if key not in ["commands", "tags"] %}
  {{ key }} = "{{ value }}"
      {% endif %}
  {% endfor %}
  {% if cmd.tags %}
  [inputs.exec.tags]
    {% for tag, tag_value in cmd.tags|dictsort %}
    {{ tag }} = "{{ tag_value }}"
    {% endfor %}
  {% endif %}

{% endfor %}
"""
    config_path = '{}/{}.conf'.format(get_configs_dir(), 'exec')
    commands = exec_rel.commands()
    if not commands:
        hookenv.log("No Commands defined in the exec relation, doing nothing.")
        return
    timeout_support = exec_timeout_supported()
    pre_proc_cmds = []
    for command in commands:
        if not timeout_support:
            command.pop('timeout')
        run_on_this_unit = command.pop('run_on_this_unit')
        if run_on_this_unit:
            pre_proc_cmds.append(command)
    if pre_proc_cmds:
        input_config = render_template(template, {'commands': pre_proc_cmds})
        hookenv.log("Updating {} plugin config file".format('exec'))
        host.write_file(config_path, input_config.encode('utf-8'))
        set_state('plugins.exec.configured')
    else:
        # if no commands, remove previous config
        if os.path.exists(config_path):
            os.unlink(config_path)
    set_state('telegraf.needs_reload')


@when_not('exec.available')
@when('plugins.exec.configured')
def exec_input_departed():
    config_path = '{}/{}.conf'.format(get_configs_dir(), 'exec')
    rels = hookenv.relations_of_type('exec')
    if not rels:
        remove_state('plugins.exec.configured')
        if os.path.exists(config_path):
            os.unlink(config_path)
        set_state('telegraf.needs_reload')


@when('influxdb-api.available')
def influxdb_api_output(influxdb):
    required_keys = ['hostname', 'port', 'user', 'password']
    rels = hookenv.relations_of_type('influxdb-api')
    endpoints = []
    user = None
    password = None
    for rel in rels:
        if all([rel.get(key) for key in required_keys]):
            endpoints.append("http://{}:{}".format(rel['hostname'], rel['port']))
            if user is None:
                user = rel['user']
            if password is None:
                password = rel['password']
    config_path = '{}/{}.conf'.format(get_configs_dir(), 'influxdb-api')
    if endpoints:
        hookenv.log("Updating {} plugin config file".format('influxdb-api'))
        content = render(source='influxdb-api.conf.tmpl', target=None,
                         templates_dir=get_templates_dir(),
                         context={'urls': json.dumps(endpoints),
                                  'username': '{}'.format(user),
                                  'password': '{}'.format(password)})
        extra_opts = render_extra_options("outputs", "influxdb")
        host.write_file(config_path, '\n'.join([content, extra_opts]).encode('utf-8'))
        set_state('plugins.influxdb-api.configured')
    elif os.path.exists(config_path):
        os.unlink(config_path)
    set_state('telegraf.needs_reload')


@when('prometheus-client.available')
def prometheus_client(prometheus):
    template = """
[[outputs.prometheus_client]]
  listen = "{{ listen }}"
"""
    if get_prometheus_port():
        hookenv.log("Prometheus configured globally, skipping plugin setup")
        prometheus.configure(get_prometheus_port())
        # bail out, nothing more need to be configured here
        return
    port = 9126
    extra_options = get_extra_options()
    options = extra_options['outputs'].get('prometheus-client', {})
    listen = options.pop('listen', None)
    if listen is not None:
        hookenv.log("Configuring prometheus_client plugin to listen on: '{}'".format(listen))
        port = int(listen.split(":", 1)[1])
    else:
        listen = ":{}".format(port)
    check_prometheus_port("prometheus_output", port)
    prometheus.configure(port)
    config_path = '{}/{}.conf'.format(get_configs_dir(), 'prometheus-client')
    hookenv.log("Updating {} plugin config file".format('prometheus-client'))
    context = {"listen": listen}
    content = render_template(template, context) + \
        render_extra_options("outputs", "prometheus_client",
                             extra_options=extra_options)
    host.write_file(config_path, content.encode('utf-8'))
    set_state('plugins.prometheus-client.configured')
    set_state('telegraf.needs_reload')


@when_not('prometheus-client.available')
@when('plugins.prometheus-client.configured')
def prometheus_client_departed():
    hookenv.log("prometheus-client relation not available")
    config_path = '{}/{}.conf'.format(get_configs_dir(), 'prometheus-client')
    rels = hookenv.relations_of_type('prometheus-client')
    if not rels and os.path.exists(config_path):
        hookenv.log("Deleting {} plugin config file".format('prometheus-client'))
        os.unlink(config_path)
        remove_state('plugins.prometheus-client.configured')
        set_state('telegraf.needs_reload')


@when('telegraf.needs_reload')
def start_or_restart():
    states = sorted([k for k in get_states().keys()
                     if k.startswith('plugins') or k.startswith('extra_plugins')])

    config_files_changed = helpers.any_file_changed(list_config_files())
    active_plugins_changed = helpers.data_changed('active_plugins', states or '')
    if (not host.service_running('telegraf') or
            config_files_changed or active_plugins_changed):
        hookenv.log("Restarting telegraf")
        host.service_restart('telegraf')
    else:
        hookenv.log("Not restarting: active_plugins_changed={} | "
                    "config_files_changed={}".format(active_plugins_changed,
                                                     config_files_changed))

    # Give telegraf time to restart.
    timeout = time.time() + 15
    while not host.service_running('telegraf') and time.time() < timeout:
        time.sleep(0.1)

    if host.service_running('telegraf'):
        hookenv.status_set('active',
                           'Monitoring {}'.format(get_remote_unit_name()))
        remove_state('telegraf.needs_reload')
    else:
        hookenv.status_set('blocked',
                           'Telegraf failed to start. Check config.')
