# Copyright 2015-2016 Canonical Ltd.
#
# This file is part of the Telegraf Charm for Juju.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from collections import namedtuple
import os
import sys
import unittest
from unittest.mock import ANY, MagicMock, patch, sentinel

# Mock layer modules
import charms
promreg = MagicMock()
charms.promreg = promreg
sys.modules['charms.promreg'] = promreg

import reactive
from reactive import telegraf


class TestPostgreSQL(unittest.TestCase):
    def test_choose_postgresql_database(self):
        m = MagicMock()
        telegraf.choose_postgresql_database(m)
        m.set_database.assert_called_once_with('postgres')

    @patch('reactive.telegraf.render_postgresql_tmpl')
    @patch('reactive.telegraf.toggle_state')
    @patch('charmhelpers.core.hookenv.unit_private_ip')
    @patch('charmhelpers.context.Relations')
    def test_postgresql_input(self, relations, unit_private_ip, toggle_state,
                              render):
        pgsql = MagicMock()
        pgsql.relation_name = 'dbrel'
        relation = MagicMock()
        relations()[pgsql.relation_name].items.return_value = [('dbrel:1',
                                                                relation)]
        relation.items.return_value = [('pg/0',
                                        {'private-address': sentinel.priv_ip,
                                         'state': 'standalone'})]
        unit_private_ip.return_value = sentinel.priv_ip

        conn_str = namedtuple('conn_str', ['host'])(sentinel.host)
        pgsql.connection_string.return_value = conn_str

        reactive.telegraf.postgresql_input(pgsql)

        render.assert_called_once_with([{'replica': 'standalone',
                                         'conn_str': conn_str,
                                         'server': sentinel.host}])
        toggle_state.assert_called_once_with('plugins.postgresql.configured',
                                             True)

    @patch('reactive.telegraf.render_extra_options')
    @patch('charmhelpers.core.host.write_file')
    @patch('reactive.telegraf.get_templates_dir')
    def test_render_postgresql_tmpl(self, get_templates_dir, write_file,
                                    extra_options):
        get_templates_dir.return_value = os.path.join(
            os.path.dirname(__file__), '..', 'templates')
        extra_options.return_value = ''

        reactive.telegraf.render_postgresql_tmpl([
            {'replica': 'master',
             'conn_str': 'my_conn_str',
             'server': 'my_server'},
        ])

        write_file.assert_called_once_with(
            '/etc/telegraf/telegraf.d/postgresql.conf', ANY
        )

    @patch('os.unlink')
    @patch('os.path.exists')
    def test_render_postgresql_tmpl_no_context(self, exists, unlink):
        exists.return_value = False
        reactive.telegraf.render_postgresql_tmpl([])
        exists.assert_called_once_with(
            '/etc/telegraf/telegraf.d/postgresql.conf'
        )
        self.assertFalse(unlink.called)
        exists.return_value = True
        reactive.telegraf.render_postgresql_tmpl([])
        unlink.assert_called_once_with(
            '/etc/telegraf/telegraf.d/postgresql.conf'
        )
